from functools import partial
from typing import Optional, Union

import numpy as np
import xarray as xr
from scipy.signal import convolve, fftconvolve

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Clip, Number
from starfish.core.util import click
from ._base import FilterAlgorithmBase
from .util import (
    determine_axes_to_group_by,
    gaussian_kernel,
)


class DeconvolvePSF(FilterAlgorithmBase):
    """
    Deconvolve a point spread function from the image. The point spread function is assumed to be
    an isotropic Gaussian, with a user specified standard deviation, sigma.

    There are currently several issues with this function.
    See https://github.com/spacetx/starfish/issues/731

    Parameters
    ----------
    num_iter : int
        number of iterations to run. Note that this is a very important parameter that requires
        careful optimization
    sigma : Number
        standard deviation of the gaussian kernel used to construct the point spread function
    is_volume: bool
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.
    clip_method : Union[str, Clip]
        (Default Clip.CLIP) Controls the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].
        Clip.CLIP: data above 1 are set to 1, and below 0 are set to 0
        Clip.SCALE_BY_IMAGE: data above 1 are scaled by the maximum value, with the maximum
        value calculated over the entire ImageStack
        Clip.SCALE_BY_CHUNK: data above 1 are scaled by the maximum value, with the maximum
        value calculated over each slice, where slice shapes are determined by the group_by
        parameters

    Examples
    --------
    >>> from skimage import color, data, restoration
    >>> camera = color.rgb2gray(data.camera())
    >>> from scipy.signal import convolve2d
    >>> psf = np.ones((5, 5)) / 25
    >>> camera = convolve2d(camera, psf, 'same')
    >>> camera += 0.1 * camera.std() * np.random.standard_normal(camera.shape)
    >>> deconvolved = restoration.richardson_lucy(camera, psf, 5)

    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/Richardson%E2%80%93Lucy_deconvolution

    Notes
    -----
    This code is based on code from skimage.restoration. We modified it to implement a bugfix
    wherein zeros in the input image or zeros produced during an intermediate would induce
    divide by zero -> Nan. These Nans would then propagate throughout the image, invalidating
    the results. Longer term, we will make a PR to skimage to introduce the fix. There is some
    existing work linked here: https://github.com/scikit-image/scikit-image/issues/2551

    """

    def __init__(
        self, num_iter: int, sigma: Number, is_volume: bool = False,
        clip_method: Union[str, Clip] = Clip.CLIP
    ) -> None:

        self.num_iter = num_iter
        self.sigma = sigma
        self.kernel_size: int = int(2 * np.ceil(2 * sigma) + 1)
        self.psf: np.ndarray = gaussian_kernel(
            shape=(self.kernel_size, self.kernel_size),
            sigma=sigma
        )
        self.is_volume = is_volume
        self.clip_method = clip_method

    _DEFAULT_TESTING_PARAMETERS = {"num_iter": 2, "sigma": 1}

    # Here be dragons. This algorithm had a bug, but the results looked nice. Now we've "fixed" it
    # and the results look bad. #548 addresses this problem.
    @staticmethod
    def _richardson_lucy_deconv(
            image: Union[xr.DataArray, np.ndarray], iterations: int, psf: np.ndarray
    ) -> np.ndarray:
        """
        Deconvolves input image with a specified point spread function.

        Parameters
        ----------
        image : Union[xr.DataArray, np.ndarray]
           Input degraded image (can be N dimensional).
        psf : ndarray
           The point spread function.
        iterations : int
           Number of iterations. This parameter plays the role of
           regularisation.

        Returns
        -------
        im_deconv : ndarray
           The deconvolved image.

        """
        # compute the times for direct convolution and the fft method. The fft is of
        # complexity O(N log(N)) for each dimension and the direct method does
        # straight arithmetic (and is O(n*k) to add n elements k times)
        direct_time = np.prod(image.shape + psf.shape)
        fft_time = np.sum([n * np.log(n) for n in image.shape + psf.shape])

        # see whether the fourier transform convolution method or the direct
        # convolution method is faster (discussed in scikit-image PR #1792)
        time_ratio = 40.032 * fft_time / direct_time

        if time_ratio <= 1 or len(image.shape) > 2:
            convolve_method = fftconvolve
        else:
            convolve_method = convolve

        image = image.astype(np.float)
        psf = psf.astype(np.float)
        im_deconv = 0.5 * np.ones(image.shape)
        psf_mirror = psf[::-1, ::-1]

        eps = np.finfo(image.dtype).eps
        for _ in range(iterations):
            x = convolve_method(im_deconv, psf, 'same')
            np.place(x, x == 0, eps)
            relative_blur = image / x + eps
            im_deconv *= convolve_method(relative_blur, psf_mirror, 'same')

        if np.all(np.isnan(im_deconv)):
            raise RuntimeError(
                'All-NaN output data detected. Likely cause is that deconvolution has been run for '
                'too many iterations.')

        return im_deconv

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose=False,
            n_processes: Optional[int] = None,
            *args,
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on the percentage completed during processing (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        func = partial(
            self._richardson_lucy_deconv,
            iterations=self.num_iter, psf=self.psf
        )
        result = stack.apply(
            func,
            group_by=group_by,
            verbose=verbose,
            n_processes=n_processes,
            in_place=in_place,
        )
        return result

    @staticmethod
    @click.command("DeconvolvePSF")
    @click.option(
        '--num-iter', type=int, help='number of iterations to run')
    @click.option(
        '--sigma', type=float, help='standard deviation of gaussian kernel')
    @click.option("--is-volume", is_flag=True,
                  help="indicates that the image stack should be filtered in 3d")
    @click.option(
        "--clip-method", default=Clip.CLIP, type=Clip,
        help="method to constrain data to [0,1]. options: 'clip', 'scale_by_image', "
             "'scale_by_chunk'")
    @click.pass_context
    def _cli(ctx, num_iter, sigma, is_volume, clip_method):
        ctx.obj["component"]._cli_run(ctx, DeconvolvePSF(num_iter, sigma, is_volume, clip_method))



from functools import partial
from typing import Optional, Union

import numpy as np
import xarray as xr
from trackpy import bandpass

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Clip, Number
from starfish.core.util import click
from ._base import FilterAlgorithmBase
from .util import determine_axes_to_group_by


class Bandpass(FilterAlgorithmBase):
    """
    Convolve with a Gaussian to remove short-wavelength noise and subtract out long-wavelength
    variations, retaining features of intermediate scale. This implementation relies on
    scipy.ndimage.filters.gaussian_filter.

    This method is a thin wrapper around :code:`trackpy.Bandpass`:

    http://soft-matter.github.io/trackpy/v0.3.0/generated/trackpy.bandpass.html

    Parameters
    ----------
    lshort : float
        filter frequencies below this value
    llong : int
        filter frequencies above this odd integer value
    threshold : float
        zero any spots below this intensity value after background subtraction (default 0)
    truncate : float
        truncate the gaussian kernel, used by the gaussian filter, at this many standard
        deviations (default 4)
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) planes
    clip_method : Union[str, Clip]
        (Default Clip.CLIP) Controls the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].
        Clip.CLIP: data above 1 are set to 1, and below 0 are set to 0
        Clip.SCALE_BY_IMAGE: data above 1 are scaled by the maximum value, with the maximum
        value calculated over the entire ImageStack
        Clip.SCALE_BY_CHUNK: data above 1 are scaled by the maximum value, with the maximum
        value calculated over each slice, where slice shapes are determined by the group_by
        parameters
    """

    def __init__(
        self, lshort: Number, llong: int, threshold: Number = 0, truncate: Number = 4,
        is_volume: bool = False, clip_method: Union[str, Clip] = Clip.CLIP
    ) -> None:
        self.lshort = lshort
        self.llong = llong

        if threshold is None:
            raise ValueError("Threshold cannot be None, please pass a float or integer")

        self.threshold = threshold
        self.truncate = truncate
        self.is_volume = is_volume
        self.clip_method = clip_method

    _DEFAULT_TESTING_PARAMETERS = {"lshort": 1, "llong": 3, "threshold": 0.01}

    @staticmethod
    def _bandpass(
            image: Union[xr.DataArray, np.ndarray],
            lshort: Number, llong: int, threshold: Number, truncate: Number
    ) -> np.ndarray:
        """Apply a bandpass filter to remove noise and background variation

        Parameters
        ----------
        image : Union[xr.DataArray, np.ndarray]
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any spots below this intensity value after background subtraction (default 0)
        truncate : float
            truncate the gaussian kernel, used by the gaussian filter, at this many standard
            deviations

        Returns
        -------
        np.ndarray :
            bandpassed image

        """
        bandpassed = bandpass(
            image, lshort=lshort, llong=llong, threshold=threshold,
            truncate=truncate
        )
        return bandpassed

    def run(
            self,
            stack: ImageStack,
            in_place: bool=False,
            verbose: bool=False,
            n_processes: Optional[int]=None,
            *args,
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        bandpass_ = partial(
            self._bandpass,
            lshort=self.lshort, llong=self.llong, threshold=self.threshold, truncate=self.truncate
        )

        group_by = determine_axes_to_group_by(self.is_volume)

        result = stack.apply(
            bandpass_,
            group_by=group_by,
            in_place=in_place,
            n_processes=n_processes,
            clip_method=self.clip_method,
            verbose=verbose,
        )
        return result

    @staticmethod
    @click.command("Bandpass")
    @click.option(
        "--lshort", type=float, help="filter signals below this frequency")
    @click.option(
        "--llong", type=int, help="filter signals above this frequency")
    @click.option(
        "--threshold", default=0, type=float, help="zero pixels below this intensity value")
    @click.option(
        "--truncate", default=4, type=float,
        help="truncate the filter at this many standard deviations")
    @click.option(
        "--clip-method", default=Clip.CLIP, type=Clip,
        help="method to constrain data to [0,1]. options: 'clip', 'scale_by_image', "
             "'scale_by_chunk'")
    @click.pass_context
    def _cli(ctx, lshort, llong, threshold, truncate, clip_method):
        ctx.obj["component"]._cli_run(
            ctx,
            Bandpass(lshort, llong, threshold, truncate, clip_method)
        )



from functools import partial
from typing import Optional, Union

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Axes, Clip
from starfish.core.util import click
from ._base import FilterAlgorithmBase


class LinearUnmixing(FilterAlgorithmBase):
    """
    LinearUnmixing enables the user to correct fluorescent bleed by subtracting fractions of the
    intensities of other channels from each channel in the ImageStack.

    Examples
    --------
    The following example provides a coefficient matrix that corrects for spectral
    mixing in a 3-channel experiment.
    Channel 0 contains a mixture of itself plus 50% of the intensity of
    channel 2. Channel 1 has no mixing with other channels. Channel 3
    consists of itself plus 10% of the intensity of both channels 0 and 1.

    >>> import numpy as np
    >>> coeff_mat = np.ndarray([
    ...     [1,    0,  -0.1]
    ...     [0,    1,  -0.1]
    ...     [-0.5, 0,  1   ]
    ... ])

    The end result of this unmixing will be that 50% of channel 2 will be subtracted from
    channel 0, channel 1 will not be changed, and 10% of channels 0 and 1 will be subtracted
    from channel 2.

    Parameters
    ----------
    coeff_mat : np.ndarray
        matrix of the linear unmixing coefficients. Should take the form: B = AX, where B are
        the unmixed values, A is coeff_mat and X are the observed values. coeff_mat has shape
        (n_ch, n_ch), and poses each channel (column) as a combination of other columns (rows).
    clip_method : Union[str, Clip]
        (Default Clip.CLIP) Controls the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].
        Clip.CLIP: data above 1 are set to 1, and below 0 are set to 0
        Clip.SCALE_BY_IMAGE: data above 1 are scaled by the maximum value, with the maximum
        value calculated over the entire ImageStack
        Clip.SCALE_BY_CHUNK: data above 1 are scaled by the maximum value, with the maximum
        value calculated over each slice, where slice shapes are determined by the group_by
        parameters

    """

    def __init__(
        self, coeff_mat: np.ndarray, clip_method: Union[str, Clip] = Clip.SCALE_BY_IMAGE,
    ) -> None:

        self.coeff_mat = coeff_mat
        self.clip_method = clip_method

    _DEFAULT_TESTING_PARAMETERS = {"coeff_mat": np.array([[1, -0.25], [-0.25, 1]])}

    @staticmethod
    def _unmix(image: xr.DataArray, coeff_mat: np.ndarray) -> np.ndarray:
        """Perform linear unmixing of channels

        Parameters
        ----------
        image : np.ndarray
            image to be scaled

        coeff_mat : np.ndarray
            matrix of the linear unmixing coefficients. Should take the form:
            B = AX, where B are the unmixed values, A is coeff_mat and X are
            the observed values. coeff_mat has shape (n_ch, n_ch), and poses
            each channel (column) as a combination of other columns (rows).

        Returns
        -------
        np.ndarray :
          Numpy array of same shape as image

        """

        x = image.sizes[Axes.X.value]
        y = image.sizes[Axes.Y.value]
        c = image.sizes[Axes.CH.value]

        # broadcast each channel coefficient across x and y
        broadcast_coeff = np.tile(coeff_mat, reps=x * y).reshape(c, y, x, c)

        # multiply the image by each coefficient
        broadcast_image = image.values[..., None] * broadcast_coeff

        # collapse the unmixed result
        unmixed_image = np.sum(broadcast_image, axis=0).transpose([2, 0, 1])

        return unmixed_image

    def run(
            self,
            stack: ImageStack,
            in_place: bool=False,
            verbose: bool=False,
            n_processes: Optional[int]=None,
            *args,
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = {Axes.ROUND, Axes.ZPLANE}
        unmix = partial(self._unmix, coeff_mat=self.coeff_mat)
        result = stack.apply(
            unmix,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            clip_method=self.clip_method,
        )
        return result

    @staticmethod
    @click.command("LinearUnmixing")
    @click.option(
        "--coeff_mat", required=True, type=np.ndarray, help="linear unmixing coefficients")
    @click.option(
        "--clip-method", default='scale_by_image',
        type=click.Choice(['clip', 'scale_by_image', 'scale_by_chunk']),
        help="method to constrain data to [0,1]")
    @click.pass_context
    def _cli(ctx, coeff_mat, clip_method):
        ctx.obj["component"]._cli_run(ctx, LinearUnmixing(coeff_mat, clip_method))



import enum
from typing import Any, Callable, Type

from starfish.core.codebook.codebook import Codebook
from starfish.core.expression_matrix.expression_matrix import ExpressionMatrix
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.util.indirectfile import (
    convert,
    GetCodebook,
    GetCodebookFromExperiment,
    GetImageStack,
    GetImageStackFromExperiment,
)


def imagestack_convert(indirect_path_or_url: str) -> ImageStack:
    """Converts a path or URL to an ImageStack.  This supports the indirect syntax, where a user
    provides a string like @<url_or_path_of_experiment.json>[fov_name][image_name].  If the indirect
    syntax is used, the experiment.json is automatically fetched and traversed to find the specified
    image in the specified field of view."""
    return convert(
        indirect_path_or_url,
        [
            GetImageStack(),
            GetImageStackFromExperiment(),
        ],
    )


def codebook_convert(indirect_path_or_url: str) -> Codebook:
    """Converts a path or URL to a Codebook.  This supports the indirect syntax, where a user
    provides a string like @<url_or_path_of_experiment.json>.  If the indirect syntax is used, the
    experiment.json is automatically fetched to find the codebook."""
    return convert(
        indirect_path_or_url,
        [
            GetCodebook(),
            GetCodebookFromExperiment(),
        ],
    )


class FileTypes(enum.Enum):
    """These are the filetypes supported as inputs and outputs for recipes.  Each filetype is
    associated with the implementing class, the method to invoke to load such a filetype, and the
    method to invoke to save back to the filetype.

    The load method is expected to be called with a string, which is the file or url to load from,
    and is expected to return an instantiated object.

    The save method is expected to be called with the object and a string, which is the path to
    write the object to.
    """
    IMAGESTACK = (ImageStack, imagestack_convert, ImageStack.export)
    INTENSITYTABLE = (IntensityTable, IntensityTable.open_netcdf, IntensityTable.to_netcdf)
    EXPRESSIONMATRIX = (ExpressionMatrix, ExpressionMatrix.load, ExpressionMatrix.save)
    CODEBOOK = (Codebook, codebook_convert, Codebook.to_json)

    def __init__(self, cls: Type, loader: Callable[[str], Any], saver: Callable[[Any, str], None]):
        self._cls = cls
        self._load = loader
        self._save = saver

    @property
    def load(self) -> Callable[[str], Any]:
        return self._load

    @property
    def save(self) -> Callable[[Any, str], None]:
        return self._save

    @staticmethod
    def resolve_by_class(cls: Type) -> "FileTypes":
        for member in FileTypes.__members__.values():
            if cls == member.value[0]:
                return member
        raise TypeError(f"filetype {cls} not supported.")

    @staticmethod
    def resolve_by_instance(instance) -> "FileTypes":
        for member in FileTypes.__members__.values():
            if isinstance(instance, member.value[0]):
                return member
        raise TypeError(f"filetype of {instance.__class__} not supported.")


class FileProvider:
    """This is used to wrap paths or URLs that are passed into Runnables via the `file_inputs` magic
    variable.  This is so we can differentiate between strings and `file_inputs` values, which must
    be first constructed into a starfish object via its loader."""
    def __init__(self, path_or_url: str) -> None:
        self.path_or_uri = path_or_url

    def __str__(self):
        return f"FileProvider(\"{self.path_or_uri}\")"


class TypedFileProvider:
    """Like :py:class:`FileProvider`, this is used to wrap paths or URLs that are passed into
    Runnables via the `file_inputs` magic variable.  In this case, the object type has been
    resolved by examining the type annotation."""
    def __init__(self, backing_file_provider: FileProvider, object_class: Type) -> None:
        self.backing_file_provider = backing_file_provider
        self.type = FileTypes.resolve_by_class(object_class)

    def load(self) -> Any:
        return self.type.load(self.backing_file_provider.path_or_uri)



from abc import abstractmethod
from typing import Callable, Sequence, Tuple, Type

import numpy as np

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.pipeline.pipelinecomponent import PipelineComponent
from starfish.core.types import Number
from starfish.core.util import click
from starfish.core.util.click.indirectparams import CodebookParamType, ImageStackParamType
from .combine_adjacent_features import ConnectedComponentDecodingResult


class DetectPixels(PipelineComponent):
    """
    Decode an image by first coding each pixel, then combining the results into spots.
    """
    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        image_stack: ImageStack = ctx.obj["image_stack"]
        intensities: IntensityTable
        intensities, _ = instance.run(image_stack)
        intensities.to_netcdf(output)

    @staticmethod
    @click.group("DetectPixels")
    @click.option("-i", "--input", required=True, type=ImageStackParamType)
    @click.option("-o", "--output", required=True)
    @click.option(
        "--codebook",
        default=None, required=True, type=CodebookParamType,
        help=(
            "A spaceTx spec-compliant json file that describes a three dimensional tensor "
            "whose values are the expected intensity of a spot for each code in each imaging "
            "round and each color channel."
        )
    )
    @click.pass_context
    def _cli(ctx, input, output, codebook):
        """pixel-wise spot detection and decoding"""
        print('Detecting Spots ...')
        ctx.obj = dict(
            component=DetectPixels,
            image_stack=input,
            output=output,
            codebook=codebook,
        )


class DetectPixelsAlgorithmBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return DetectPixels

    @abstractmethod
    def run(
            self,
            primary_image: ImageStack,
            *args,
    ) -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
        """Finds spots in an ImageStack"""
        raise NotImplementedError()

    @staticmethod
    def _get_measurement_function(measurement_type: str) -> Callable[[Sequence], Number]:
        try:
            measurement_function = getattr(np, measurement_type)
        except AttributeError:
            raise ValueError(
                f'measurement_type must be a numpy reduce function such as "max" or "mean". '
                f'{measurement_type} not found.')
        return measurement_function



import os
import tempfile

import pandas as pd
import pytest

from starfish import IntensityTable
from starfish.core.codebook.test.factories import codebook_array_factory
from starfish.core.types import Coordinates, DecodedSpots, Features


def dummy_intensities() -> IntensityTable:

    codebook = codebook_array_factory()
    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=10,
        height=10,
        width=10,
        n_spots=5,
    )

    intensities[Coordinates.Z.value] = (Features.AXIS, [0, 1, 0, 1, 0])
    intensities[Coordinates.Y.value] = (Features.AXIS, [10, 30, 50, 40, 20])
    intensities[Coordinates.X.value] = (Features.AXIS, [50.2, 30.2, 60.2, 40.2, 70.2])

    # remove target from dummy to test error messages
    del intensities[Features.TARGET]

    return intensities


def test_decoded_spots() -> None:
    data = dummy_intensities()

    with pytest.raises(RuntimeError):
        data.to_decoded_spots()

    # mock decoder run by adding target list
    data[Features.TARGET] = (Features.AXIS, list('abcde'))

    ds = data.to_decoded_spots()

    assert ds.data.shape[0] == 5

    # test serialization
    tempdir = tempfile.mkdtemp()
    filename = os.path.join(tempdir, 'spots.csv')
    ds.save_csv(filename)

    # load back into memory
    ds2 = DecodedSpots.load_csv(filename)
    pd.testing.assert_frame_equal(ds.data, ds2.data)



#!/usr/bin/env python

import os
import setuptools
import versioneer

install_requires = [
    line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "REQUIREMENTS.txt"))
]

setuptools.setup(
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    extras_require={
        'napari': ['napari==0.0.6'],
    },
    entry_points={
        'console_scripts': [
            "starfish=starfish:starfish",
        ]
    },
    include_package_data=True,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



from typing import Optional, Union

import numpy as np
import xarray as xr
from skimage.morphology import ball, disk, white_tophat

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Clip
from starfish.core.util import click
from ._base import FilterAlgorithmBase
from .util import determine_axes_to_group_by


class WhiteTophat(FilterAlgorithmBase):
    """
    Performs "white top hat" filtering of an image to enhance spots. White top hat filtering
    finds spots that are both smaller and brighter than their surroundings by subtracting an
    estimate of the background produced by a binary opening of the image using a disk-shaped
    structuring element.

    Parameters
    ----------
    masking_radius : int
        radius of the morphological masking structure in pixels
    is_volume : int
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.
    clip_method : Union[str, Clip]
        (Default Clip.CLIP) Controls the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].
        Clip.CLIP: data above 1 are set to 1, and below 0 are set to 0
        Clip.SCALE_BY_IMAGE: data above 1 are scaled by the maximum value, with the maximum
        value calculated over the entire ImageStack
        Clip.SCALE_BY_CHUNK: data above 1 are scaled by the maximum value, with the maximum
        value calculated over each slice, where slice shapes are determined by the group_by
        parameters

    Notes
    -----
    See https://en.wikipedia.org/wiki/Top-hat_transform for more information

    """

    def __init__(
        self, masking_radius: int, is_volume: bool = False,
        clip_method: Union[str, Clip] = Clip.CLIP
    ) -> None:

        self.masking_radius = masking_radius
        self.is_volume = is_volume
        self.clip_method = clip_method

    _DEFAULT_TESTING_PARAMETERS = {"masking_radius": 3}

    def _white_tophat(self, image: Union[xr.DataArray, np.ndarray]) -> np.ndarray:
        if self.is_volume:
            structuring_element = ball(self.masking_radius)
        else:
            structuring_element = disk(self.masking_radius)
        return white_tophat(image, selem=structuring_element)

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        result = stack.apply(
            self._white_tophat,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            clip_method=self.clip_method
        )
        return result

    @staticmethod
    @click.command("WhiteTophat")
    @click.option(
        "--masking-radius", default=15, type=int,
        help="diameter of morphological masking disk in pixels")
    @click.option(  # FIXME: was this intentionally missed?
        "--is-volume", is_flag=True, help="filter 3D volumes")
    @click.option(
        "--clip-method", default=Clip.CLIP, type=Clip,
        help="method to constrain data to [0,1]. options: 'clip', 'scale_by_image', "
             "'scale_by_chunk'")
    @click.pass_context
    def _cli(ctx, masking_radius, is_volume, clip_method):
        ctx.obj["component"]._cli_run(ctx, WhiteTophat(masking_radius, is_volume, clip_method))



"""
Note: this blob detector uses the same underlying methods (skimage.blob_log, skimage.blob_dog) as
the starfish.spots.SpotFinder.BlobDetector. In the future, this and the packages other blob
detectors should be refactored to merge their functionalities.

See: https://github.com/spacetx/starfish/issues/1005
"""

from collections import defaultdict
from typing import Dict, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr
from click import Choice
from sklearn.neighbors import NearestNeighbors

from starfish.core.compat import blob_dog, blob_log
from starfish.core.image._filter.util import determine_axes_to_group_by
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.intensity_table.intensity_table_coordinates import \
    transfer_physical_coords_from_imagestack_to_intensity_table
from starfish.core.types import Axes, Features, Number, SpotAttributes
from starfish.core.util import click
from ._base import DetectSpotsAlgorithmBase

blob_detectors = {
    'blob_dog': blob_dog,
    'blob_log': blob_log
}


class LocalSearchBlobDetector(DetectSpotsAlgorithmBase):
    """
    Multi-dimensional gaussian spot detector.

    This method is a wrapper for skimage.feature.blob_log that engages in a local radius search
    to match spots up across rounds. In sparse conditions, the search can be seeded in one round
    with a relatively large radius. In crowded images, the spots can be seeded in all rounds and
    consensus filtering can be used to extract codes that are consistently extracted across
    rounds.

    Parameters
    ----------
    min_sigma : float
        The minimum standard deviation for Gaussian Kernel. Keep this low to
        detect smaller blobs.
    max_sigma : float
        The maximum standard deviation for Gaussian Kernel. Keep this high to
        detect larger blobs.
    threshold : float
        The absolute lower bound for scale space maxima. Local maxima smaller
        than thresh are ignored. Reduce this to detect blobs with less
        intensities.
    detector_method : str ['blob_dog', 'blob_log']
        Name of the type of detection method used from skimage.feature, default: blob_log.
    search_radius : int
        Number of pixels over which to search for spots in other rounds and channels.
    anchor_round : int
        The imaging round against which other rounds will be checked for spots in the same
        approximate pixel location.
    detector_kwargs : Dict[str, Any]
        Additional keyword arguments to pass to the detector_method.

    """

    def __init__(
            self,
            min_sigma: Union[Number, Tuple[Number, ...]],
            max_sigma: Union[Number, Tuple[Number, ...]],
            num_sigma: int,
            threshold: Number,
            detector_method: str='blob_log',
            exclude_border: Optional[int]=None,
            search_radius: int=3,
            anchor_round: int=1,
            **detector_kwargs,
    ) -> None:

        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.threshold = threshold
        self.is_volume = True  # TODO test 2-d spot calling
        self.exclude_border = exclude_border
        self.search_radius = search_radius
        self.anchor_round = anchor_round
        self.detector_kwargs = detector_kwargs
        try:
            self.detector_method = blob_detectors[detector_method]
        except ValueError:
            raise ValueError(f"Detector method must be one of {list(blob_detectors.keys())}")

    def _spot_finder(self, data: xr.DataArray) -> pd.DataFrame:
        """Find spots in a data volume.

        Parameters
        ----------
        data : xr.DataArray
            3D (z, y, x) data volume in which spots will be identified.

        Returns
        -------
        pd.DataFrame
            DataFrame wrapping the output of skimage blob_log or blob_dog with named outputs.

        """

        # results = np.ndarray: n_spots x (z, y, x, radius, sigma_z, sigma_y, sigma_x)
        results = self.detector_method(
            data,
            min_sigma=self.min_sigma,
            max_sigma=self.max_sigma,
            threshold=self.threshold,
            exclude_border=self.exclude_border,
            **self.detector_kwargs
        )

        # if spots were detected
        if results.shape[0]:

            # measure intensities
            z_inds = results[:, 0].astype(int)
            y_inds = results[:, 1].astype(int)
            x_inds = results[:, 2].astype(int)
            intensities = data.values[tuple([z_inds, y_inds, x_inds])]

            # collapse radius if sigma is non-scalar
            if all(np.isscalar(s) for s in (self.min_sigma, self.max_sigma)):
                radius = results[:, 3]
            else:
                radius = np.mean(results[:, -3:], axis=1)

            # construct dataframe
            spot_data = pd.DataFrame(
                data={
                    "intensity": intensities,
                    Axes.ZPLANE: z_inds,
                    Axes.Y: y_inds,
                    Axes.X: x_inds,
                    Features.SPOT_RADIUS: radius
                }
            )

        else:
            spot_data = pd.DataFrame(
                data=np.array(
                    [],
                    dtype=[
                        ('intensity', float), ('z', int), ('y', int), ('x', int),
                        (Features.SPOT_RADIUS, float)
                    ]
                )
            )

        return spot_data

    def _find_spots(
        self,
        data_stack: ImageStack,
        verbose: bool=False,
        n_processes: Optional[int]=None
    ) -> Dict[Tuple[int, int], np.ndarray]:
        """Find spots in all (z, y, x) volumes of an ImageStack.

        Parameters
        ----------
        data_stack : ImageStack
            Stack containing spots to find.

        Returns
        -------
        Dict[Tuple[int, int], np.ndarray]
            Dictionary mapping (round, channel) pairs to a spot table generated by skimage blob_log
            or blob_dog.

        """
        # find spots in each (r, c) volume
        transform_results = data_stack.transform(
            self._spot_finder,
            group_by=determine_axes_to_group_by(self.is_volume),
            n_processes=n_processes,
        )

        # create output dictionary
        spot_results = {}
        for spot_calls, axes_dict in transform_results:
            r = axes_dict[Axes.ROUND]
            c = axes_dict[Axes.CH]
            spot_results[r, c] = spot_calls

        return spot_results

    @staticmethod
    def _merge_spots_by_round(
        spot_results: Dict[Tuple[int, int], pd.DataFrame]
    ) -> Dict[int, pd.DataFrame]:
        """Merge DataFrames containing spots from different channels into one DataFrame per round.

        Executed on the output of
        Parameters
        ----------
        spot_results : Dict[Tuple[int, int], pd.DataFrame]
            Output of _find_spots. Dictionary mapping (round, channel) volumes to the spots detected
            in them.

        Returns
        -------
        Dict[int, pd.DataFrame]
            Dictionary mapping round volumes to the spots detected in them. Contains an additional
            column labeled by Axes.CH which identifies the channel in which a given spot was
            detected.

        """

        # add channel information to each table and add it to round_data
        round_data: Mapping[int, List] = defaultdict(list)
        for (r, c), df in spot_results.items():
            df[Axes.CH.value] = np.full(df.shape[0], c)
            round_data[r].append(df)

        # create one dataframe per round
        round_dataframes = {
            k: pd.concat(v, axis=0).reset_index().drop('index', axis=1)
            for k, v in round_data.items()
        }

        return round_dataframes

    @staticmethod
    def _match_spots(
        round_dataframes: Dict[int, pd.DataFrame], search_radius: int, anchor_round: int
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """ For each spot in anchor round, find the closest spot within search_radius in all rounds.

        Parameters
        ----------
        round_dataframes : Dict[int, pd.DataFrame]
            Output from _merge_spots_by_round, contains mapping of image volumes from each round to
            all the spots detected in them.
        search_radius : int
            The maximum (euclidean) distance in pixels for a spot to be considered matching in
            a round subsequent to the anchor round.
        anchor_round : int
            The imaging round to seed the local search from.

        Returns
        -------
        pd.DataFrame
            Spots x rounds dataframe containing the distances to the nearest spot. np.nan if
            no spot is detected within search radius
        pd.DataFrame
            Spots x rounds dataframe containing the indices of the nearest spot to the
            corresponding round_dataframe. np.nan if no spot is detected within search radius.

        """
        reference_df = round_dataframes[anchor_round]
        reference_coordinates = reference_df[[Axes.ZPLANE, Axes.Y, Axes.X]]

        dist = pd.DataFrame(
            data=np.zeros((reference_df.shape[0], len(round_dataframes)), dtype=float),
            columns=list(round_dataframes.keys())
        )
        ind = pd.DataFrame(
            data=np.zeros((reference_df.shape[0], len(round_dataframes)), dtype=np.int32),
            columns=list(round_dataframes.keys())
        )

        # fill data for anchor round; every spot is a perfect match to itself.
        ind[anchor_round] = np.arange(reference_df.shape[0], dtype=np.int32)

        # get spots matching across rounds
        for r in sorted(set(round_dataframes.keys()) - {anchor_round, }):
            query_df = round_dataframes[r]
            query_coordinates = query_df[[Axes.ZPLANE, Axes.Y, Axes.X]]

            # Build the classifier; chose NN over radius neighbors because data structures are
            # amenable to vectorization, which improves execution time.
            # TODO ambrosejcarr use n_neighbors > 1 to break ties, enable codebook-based finding
            #      use additional axes in dist, ind to retain vectorization.
            nn = NearestNeighbors(n_neighbors=1)
            nn.fit(query_coordinates)

            distances, indices = nn.kneighbors(reference_coordinates)
            dist[r] = distances
            ind[r] = indices

        return dist, ind

    @staticmethod
    def _build_intensity_table(
        round_dataframes: Dict[int, pd.DataFrame],
        dist: pd.DataFrame,
        ind: pd.DataFrame,
        channels: Sequence[int],
        rounds: Sequence[int],
        search_radius: int,
        anchor_round: int,
    ) -> IntensityTable:
        """Construct an intensity table from the results of a local search over detected spots

        Parameters
        ----------
        round_dataframes : Dict[int, pd.DataFrame]
            Output from _merge_spots_by_round, contains mapping of image volumes from each round to
            all the spots detected in them.
        dist, ind : pd.DataFrame
            Output from _match_spots, contains distances and indices to the nearest spot for each
            spot in anchor_round.
        channels, rounds : Sequence[int]
            Channels and rounds present in the ImageStack from which spots were detected.
        search_radius : int
            The maximum (euclidean) distance in pixels for a spot to be considered matching in
            a round subsequent to the anchor round.
        anchor_round : int
            The imaging round to seed the local search from.

        """

        anchor_df = round_dataframes[anchor_round]

        # create empty IntensityTable filled with np.nan
        data = np.full((dist.shape[0], len(channels), len(rounds)), fill_value=np.nan)
        dims = (Features.AXIS, Axes.CH.value, Axes.ROUND.value)
        coords = {
            Features.SPOT_RADIUS: (Features.AXIS, anchor_df[Features.SPOT_RADIUS]),
            Axes.ZPLANE.value: (Features.AXIS, anchor_df[Axes.ZPLANE]),
            Axes.Y.value: (Features.AXIS, anchor_df[Axes.Y]),
            Axes.X.value: (Features.AXIS, anchor_df[Axes.X]),
            Axes.ROUND.value: (Axes.ROUND.value, rounds),
            Axes.CH.value: (Axes.CH.value, channels)
        }
        intensity_table = IntensityTable(data=data, dims=dims, coords=coords)

        # fill IntensityTable
        for r in rounds:

            # get intensity data and indices
            spot_indices = ind[r]
            intensity_data = round_dataframes[r].loc[spot_indices, 'intensity']
            channel_index = round_dataframes[r].loc[spot_indices, Axes.CH]
            round_index = np.full(ind.shape[0], fill_value=r, dtype=int)
            feature_index = np.arange(ind.shape[0], dtype=int)

            # mask spots that are outside the search radius
            mask = np.asarray(dist[r] < search_radius)  # indices need not match
            feature_index = feature_index[mask]
            channel_index = channel_index[mask]
            round_index = round_index[mask]
            intensity_data = intensity_data[mask]

            # need numpy indexing to set values in vectorized manner
            intensity_table.values[feature_index, channel_index, round_index] = intensity_data

        return intensity_table

    def run(
            self,
            primary_image: ImageStack,
            blobs_image: Optional[ImageStack] = None,
            blobs_axes: Optional[Tuple[Axes, ...]] = None,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> IntensityTable:
        """Find 1-hot coded spots in data.

        Parameters
        ----------
        data : ImageStack
            Image data containing coded spots.
        verbose : bool
            If True, report on progress of spot finding.
        n_processes : Optional[int]
            Number of processes to devote to spot finding. If None, will use the number of available
            cpus (Default None).

        Notes
        -----
        blobs_image is an unused parameter that is included for testing purposes. It should not
        be passed to this method. If it is passed, the method will trigger a ValueError.

        Returns
        -------
        IntensityTable
            Contains detected coded spots.

        """

        if blobs_image is not None:
            raise ValueError(
                "blobs_image shouldn't be set for LocalSearchBlobDetector.  This is likely a usage "
                "error."
            )

        per_tile_spot_results = self._find_spots(
            primary_image, verbose=verbose, n_processes=n_processes)

        per_round_spot_results = self._merge_spots_by_round(per_tile_spot_results)

        distances, indices = self._match_spots(
            per_round_spot_results,
            search_radius=self.search_radius, anchor_round=self.anchor_round
        )

        # TODO implement consensus seeding (SeqFISH)

        intensity_table = self._build_intensity_table(
            per_round_spot_results, distances, indices,
            rounds=primary_image.xarray[Axes.ROUND.value].values,
            channels=primary_image.xarray[Axes.CH.value].values,
            search_radius=self.search_radius,
            anchor_round=self.anchor_round
        )

        transfer_physical_coords_from_imagestack_to_intensity_table(
            image_stack=primary_image, intensity_table=intensity_table
        )

        return intensity_table

    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        # LocalSearchBlobDetector does not follow the same contract as the remaining spot detectors.
        # TODO: (ambrosejcarr) Rationalize the spot detectors by contract and then remove this hack.
        raise NotImplementedError()

    @staticmethod
    @click.command("LocalSearchBlobDetector")
    @click.option(
        "--min-sigma", default=4, type=int, help="Minimum spot size (in standard deviation).")
    @click.option(
        "--max-sigma", default=6, type=int, help="Maximum spot size (in standard deviation).")
    @click.option(
        "--threshold", default=.01, type=float, help="Dots threshold.")
    @click.option(
        "--overlap", default=0.5, type=float,
        help="Dots with overlap of greater than this fraction are combined.")
    @click.option(
        "--detector-method", default='blob_log', type=Choice(['blob_log', 'blob_dog']),
        help="Name of the type of the skimage blob detection method.")
    @click.option(
        "--search-radius", default=3, type=int,
        help="Number of pixels over which to search for spots in other image tiles.")
    @click.pass_context
    def _cli(
        ctx, min_sigma, max_sigma, threshold, overlap, show, detector_method, search_radius
    ) -> None:
        instance = LocalSearchBlobDetector(
            min_sigma, max_sigma, threshold, overlap,
            detector_method=detector_method, search_radius=search_radius
        )
        ctx.obj["component"]._cli_run(ctx, instance)



import abc
from typing import Generic, Iterable, TypeVar

from starfish.core.codebook.codebook import Codebook
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.util.indirectfile import (
    ConversionFormula,
    convert,
    GetCodebook,
    GetCodebookFromExperiment,
    GetImageStack,
    GetImageStackFromExperiment,
    NoApplicableConversionFormulaError,
    NoSuccessfulConversionFormulaError,
)
from . import ParamType


IndirectResultType = TypeVar("IndirectResultType")


class IndirectFile(ParamType, Generic[IndirectResultType]):
    def convert(self, value: str, param, ctx):
        conversion_formulas = self.get_conversion_formulas()
        try:
            return convert(value, conversion_formulas)
        except (NoApplicableConversionFormulaError, NoSuccessfulConversionFormulaError) as ex:
            self.fail(ex.args[0])

    @abc.abstractmethod
    def get_conversion_formulas(self) -> Iterable[ConversionFormula[IndirectResultType]]:
        """Return one or more conversion Formulas to get from an input string to the type of object
        we want.
        """
        raise NotImplementedError()


class CodebookParam(IndirectFile[Codebook]):
    def __init__(self):
        self.name = "codebook"

    def get_conversion_formulas(self) -> Iterable[ConversionFormula[Codebook]]:
        return [
            GetCodebookFromExperiment(),
            GetCodebook(),
        ]


CodebookParamType = CodebookParam()


class ImageStackParam(IndirectFile[ImageStack]):
    def __init__(self):
        self.name = "imagestack"

    def get_conversion_formulas(self) -> Iterable[ConversionFormula[ImageStack]]:
        return [
            GetImageStackFromExperiment(),
            GetImageStack(),
        ]


ImageStackParamType = ImageStackParam()



from .from_codebook import create_imagestack_from_codebook
from .synthetic_stack import synthetic_stack
from .unique_tiles import unique_tiles_imagestack
from .with_coords import imagestack_with_coords_factory



from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr
from skimage.feature import blob_dog, blob_doh, blob_log

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import Axes, Features, Number, SpotAttributes
from starfish.core.util import click
from ._base import DetectSpotsAlgorithmBase
from .detect import detect_spots, measure_spot_intensity

blob_detectors = {
    'blob_dog': blob_dog,
    'blob_doh': blob_doh,
    'blob_log': blob_log
}


class BlobDetector(DetectSpotsAlgorithmBase):
    """
    Multi-dimensional gaussian spot detector

    This method is a wrapper for skimage.feature.blob_log

    Parameters
    ----------
    min_sigma : float
        The minimum standard deviation for Gaussian Kernel. Keep this low to
        detect smaller blobs.
    max_sigma : float
        The maximum standard deviation for Gaussian Kernel. Keep this high to
        detect larger blobs.
    num_sigma : int
        The number of intermediate values of standard deviations to consider
        between `min_sigma` and `max_sigma`.
    threshold : float
        The absolute lower bound for scale space maxima. Local maxima smaller
        than thresh are ignored. Reduce this to detect blobs with less
        intensities.
    overlap : float [0, 1]
        If two spots have more than this fraction of overlap, the spots are combined
        (default = 0.5)
    measurement_type : str ['max', 'mean']
        name of the function used to calculate the intensity for each identified spot area
    detector_method: str ['blob_dog', 'blob_doh', 'blob_log']
        name of the type of detection method used from skimage.feature, default: blob_log

    Notes
    -----
    see also: http://scikit-image.org/docs/dev/auto_examples/features_detection/plot_blob.html

    """

    def __init__(
            self,
            min_sigma: Number,
            max_sigma: Number,
            num_sigma: int,
            threshold: Number,
            overlap: float = 0.5,
            measurement_type='max',
            is_volume: bool = True,
            detector_method: str = 'blob_log'
    ) -> None:

        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.num_sigma = num_sigma
        self.threshold = threshold
        self.overlap = overlap
        self.is_volume = is_volume
        self.measurement_function = self._get_measurement_function(measurement_type)
        try:
            self.detector_method = blob_detectors[detector_method]
        except ValueError:
            raise ValueError("Detector method must be one of {blob_log, blob_dog, blob_doh}")

    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        """
        Find spots using a gaussian blob finding algorithm

        Parameters
        ----------
        data_image : Union[np.ndarray, xr.DataArray]
            image containing spots to be detected

        Returns
        -------
        SpotAttributes :
            DataFrame of metadata containing the coordinates, intensity and radius of each spot

        """

        fitted_blobs_array: np.ndarray = self.detector_method(
            data_image,
            self.min_sigma,
            self.max_sigma,
            self.num_sigma,
            self.threshold,
            self.overlap
        )

        if fitted_blobs_array.shape[0] == 0:
            return SpotAttributes.empty(extra_fields=['intensity', 'spot_id'])

        # create the SpotAttributes Table
        columns = [Axes.ZPLANE.value, Axes.Y.value, Axes.X.value, Features.SPOT_RADIUS]
        fitted_blobs = pd.DataFrame(data=fitted_blobs_array, columns=columns)

        # convert standard deviation of gaussian kernel used to identify spot to radius of spot
        converted_radius = np.round(fitted_blobs[Features.SPOT_RADIUS] * np.sqrt(3))
        fitted_blobs[Features.SPOT_RADIUS] = converted_radius

        # convert the array to int so it can be used to index
        spots = SpotAttributes(fitted_blobs)

        spots.data['intensity'] = measure_spot_intensity(
            data_image, spots, self.measurement_function)
        spots.data['spot_id'] = np.arange(spots.data.shape[0])

        return spots

    def run(
            self,
            primary_image: ImageStack,
            blobs_image: Optional[ImageStack] = None,
            blobs_axes: Optional[Tuple[Axes, ...]] = None,
            n_processes: Optional[int] = None,
            *args,
    ) -> IntensityTable:
        """
        Find spots.

        Parameters
        ----------
        primary_image : ImageStack
            ImageStack where we find the spots in.
        blobs_image : Optional[ImageStack]
            If provided, spots will be found in the blobs image, and intensities will be measured
            across rounds and channels. Otherwise, spots are measured independently for each channel
            and round.
        blobs_axes : Optional[Tuple[Axes, ...]]
            If blobs_image is provided, blobs_axes must be provided as well.  blobs_axes represents
            the axes across which the blobs image is max projected before spot detection is done.
        n_processes : Optional[int] = None,
            Number of processes to devote to spot finding.
        """

        intensity_table = detect_spots(
            data_stack=primary_image,
            spot_finding_method=self.image_to_spots,
            reference_image=blobs_image,
            reference_image_max_projection_axes=blobs_axes,
            measurement_function=self.measurement_function,
            n_processes=n_processes,
            radius_is_gyration=False)

        return intensity_table

    @staticmethod
    @click.command("BlobDetector")
    @click.option(
        "--min-sigma", default=4, type=int, help="Minimum spot size (in standard deviation)")
    @click.option(
        "--max-sigma", default=6, type=int, help="Maximum spot size (in standard deviation)")
    @click.option(
        "--num-sigma", default=20, type=int, help="Number of sigmas to try")
    @click.option(
        "--threshold", default=.01, type=float, help="Dots threshold")
    @click.option(
        "--overlap", default=0.5, type=float,
        help="dots with overlap of greater than this fraction are combined")
    @click.option(
        "--show", default=False, is_flag=True, help="display results visually")
    @click.option(
        "--detector_method", default='blob_log',
        help="str ['blob_dog', 'blob_doh', 'blob_log'] name of the type of "
             "detection method used from skimage.feature. Default: blob_log"
    )
    @click.pass_context
    def _cli(ctx, min_sigma, max_sigma, num_sigma, threshold, overlap, show, detector_method):
        instance = BlobDetector(min_sigma, max_sigma, num_sigma, threshold, overlap,
                                detector_method=detector_method)
        #  FIXME: measurement_type, is_volume missing as options; show missing as ctor args
        ctx.obj["component"]._cli_run(ctx, instance)



"""
Tests for IntensityTable.from_spot_data method
"""

import numpy as np
import pandas as pd
import pytest

from starfish.core.types import Axes, Features, SpotAttributes
from ..intensity_table import IntensityTable


def spot_attribute_factory(n: int) -> SpotAttributes:
    """
    Construct SpotAttributes with n synthetic attributes. Each attribute has radius 1 and
    x, y, z coordinates equal to their index i in [0, n)
    """
    return SpotAttributes(
        pd.DataFrame(
            data=np.array([[i, i, i, 1] for i in np.arange(n)]),
            columns=[Axes.ZPLANE, Axes.Y, Axes.X, Features.SPOT_RADIUS]
        )
    )


def test_intensity_table_can_be_constructed_from_a_numpy_array_and_spot_attributes():
    """
    Verify that the IntensityTable can be created and that the resulting data matches the array
    it was constructed from.
    """
    spot_attributes = spot_attribute_factory(3)
    data = np.zeros(30).reshape(3, 5, 2)
    intensities = IntensityTable.from_spot_data(
        data, spot_attributes, np.arange(data.shape[1]), np.arange(data.shape[2]))

    assert intensities.shape == data.shape
    assert np.array_equal(intensities.values, data)


@pytest.mark.parametrize(
    "num_features, num_ch_values, num_round_values",
    [
        (2, 5, 2,),
        (3, 4, 2,),
        (3, 5, 1,),
    ]
)
def test_from_spot_attributes_must_have_aligned_dimensions_spot_attributes_and_data(
        num_features, num_ch_values, num_round_values,
):
    """
    Number of features must match number of SpotAttributes. Pass two attributes and 3 features and
    verify a ValueError is raised.
    """
    spot_attributes = spot_attribute_factory(num_features)
    data = np.zeros(30).reshape(3, 5, 2)
    with pytest.raises(ValueError):
        IntensityTable.from_spot_data(
            data, spot_attributes, np.arange(num_ch_values), np.arange(num_round_values))



import numpy as np

from .imagestack_test_utils import recalculate_physical_coordinate_range

COORDS = 0.01, 0.1


def test_physical_coordinate_calculator():
    # coords of the starting pixel should be (0.01, 0.01)
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, 0)
    assert np.isclose(start, 0.01)
    assert np.isclose(end, 0.01)

    # coords of the last pixel should be (0.1, 0.1)
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, 59)
    assert np.isclose(start, 0.1)
    assert np.isclose(end, 0.1)

    # the original range should be (0.01, 0.1)
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, slice(0, 60))
    assert np.isclose(start, 0.01)
    assert np.isclose(end, 0.1)

    # another way of expressing the original range
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, slice(0, -1))
    assert np.isclose(start, 0.01)
    assert np.isclose(end, 0.1)

    # yet another way of expressing the original range
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, slice(0, None))
    assert np.isclose(start, 0.01)
    assert np.isclose(end, 0.1)

    # and yet another way of expressing the original range
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, slice(None, -1))
    assert np.isclose(start, 0.01)
    assert np.isclose(end, 0.1)

    # coords of the first half should be (0.01, 0.0557627)
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, slice(0, 30))
    assert np.isclose(start, 0.01)
    assert np.isclose(end, 0.054237288135593)

    # coords of the second half should be (0.01, 0.0557627)
    start, end = recalculate_physical_coordinate_range(COORDS[0], COORDS[1], 60, slice(30, 60))
    assert np.isclose(start, 0.055762711864407)
    assert np.isclose(end, 0.1)
